(* ::Package:: *)

(* :Title: Non-Grid Interpolation *)

(* :Author: Yasushi Iwasaki *)

(* :Summary: This package provides interpolation functions of 3D data which is non-grid in domain. *)


BeginPackage["NonGridInterpolation`",{"ComputationalGeometry`"}]


Remove[NearestNeighbor]


NearestNeighbor::usage="NearestNeighbor[data3D] constructs a nearest neighbor interpolation function of data3D using boundary process function Mean.
NearestNeighbor[data3D,process] constructs a nearest neighbor interpolation function of data3D using boundary process function."

DelaunayTriangulationPiecewiseLinear::usage="DelaunayTriangulationPiecewiseLinear[data3D] constructs a Delaunay triangulation piecewise linear interpolation function of data3D using extrapolation value Null.
DelaunayTriangulationPiecewiseLinear[data3D,extrapolation] constructs a Delaunay triangulation piecewise linear interpolation function of data3D using extrapolation value or function of #1 and #2 except &."

RadialBasisFunctionNetwork::usage="RadialBasisFunctionNetwork[data3D] constructs a radial basis function network interpolation function of data3D using radial basis function Thin Plate Spline (If[#==0,0,#^2 Log[#]]&).
RadialBasisFunctionNetwork[data3D,\[Phi]] constructs a radial basis function network interpolation function of data3D using radial basis function \[Phi]."

ProcessDuplication::usage="ProcessDuplication[data3D] processes duplication point in domain of data3D using process function Mean.
ProcessDuplication[data3D,process] processes duplication point in domain of data3D using process function."


Begin["`Private`"]

NearestNeighbor::duplication=DelaunayTriangulationPiecewiseLinear::duplication=RadialBasisFunctionNetwork::duplication="There exists duplication point in domain (`1`).
Process the duplication point using ProcessDuplication.";

NearestNeighbor[data3D_,process_:Mean]:=Block[{duplicates,data2D=data3D[[All,{1,2}]],zz=data3D[[All,3]]},
If[(duplicates=Select[Tally[data2D],#[[2]]!=1&])!={},Message[NearestNeighbor::duplication,duplicates];Return[$Failed];];
Return[Evaluate[process[Quiet[Nearest[data2D->zz][{#1,#2}],{Nearest::neard}]]]&];
]

MakeTriangles[vertexInformation_]:=Block[{vertex=vertexInformation[[1]],neighbors=vertexInformation[[2]],subtenses},
If[Length[neighbors]==2,Return[{Prepend[neighbors,vertex]}]];
subtenses=Append[Partition[neighbors,2,1],#[neighbors]&/@{First,Last}];
Return[Prepend[#,vertex]&/@subtenses];
]

MakePlaneVector[triangleData3D_]:=Block[{matrix=ReplacePart[triangleData3D,{i_,3}->1],zz=triangleData3D[[All,3]]},
Return[Inverse[matrix].zz];
]

DomainQ[triangleData2D_,point_]:=Block[{matrix=Transpose[{triangleData2D[[2]]-triangleData2D[[1]],triangleData2D[[3]]-triangleData2D[[1]]}],vector=point-triangleData2D[[1]],s,t},
{s,t}=Inverse[matrix].vector;
Return[s>=0&&t>=0&&s+t<=1];
]

DelaunayTriangulationPiecewiseLinear[data3D_,extrapolation_:Null]:=Block[{duplicates,data2D=data3D[[All,{1,2}]],vertexInformations,convexHull,triangles,triangles2,delaunayTriangles,delaunayTrianglesData2D,delaunayTrianglesData3D,planeVectors},
If[(duplicates=Select[Tally[data2D],#[[2]]!=1&])!={},Message[DelaunayTriangulationPiecewiseLinear::duplication,duplicates];Return[$Failed];];
If[Length[data3D]==3,Return[Evaluate[Piecewise[{{MakePlaneVector[data3D].{#1,#2,1},DomainQ[data2D,{#1,#2}]}},extrapolation]]&]];
{vertexInformations,convexHull}=DelaunayTriangulation[data2D,Hull->True];
triangles=MakeTriangles/@vertexInformations;
triangles2=Cases[Tally[Join@@Map[Sort,triangles,2]],{triangle_,3}:>triangle];
delaunayTriangles=DeleteCases[triangles2, Sort[convexHull]];
delaunayTrianglesData2D=data2D[[#]]&/@delaunayTriangles;
delaunayTrianglesData3D=data3D[[#]]&/@delaunayTriangles;
planeVectors=MakePlaneVector/@delaunayTrianglesData3D;
Return[Evaluate[Piecewise[Table[{planeVectors[[i]].{#1,#2,1},DomainQ[delaunayTrianglesData2D[[i]],{#1,#2}]},{i,Length[delaunayTriangles]}],extrapolation]]&];
]

RadialBasisFunctionNetwork[data3D_,\[Phi]_:(If[#==0,0,#^2 Log[#]]&)]:=Block[{duplicates,n=Length[data3D],data2D=data3D[[All,{1,2}]],zz=data3D[[All,3]],\[CapitalPhi],aa},
If[(duplicates=Select[Tally[data2D],#[[2]]!=1&])!={},Message[RadialBasisFunctionNetwork::duplication,duplicates];Return[$Failed];];
\[CapitalPhi]=DiagonalMatrix[Table[\[Phi][0],{n}]];
Do[\[CapitalPhi][[i,j]]=\[CapitalPhi][[j,i]]=\[Phi][Norm[data2D[[i]]-data2D[[j]]]],{i,2,n},{j,i-1}];
aa=Inverse[\[CapitalPhi]].zz;
Return[Evaluate[aa.\[Phi]/@Table[Norm[{#1,#2}-data2D[[i]]],{i,n}]]&];
]

ProcessDuplication[data3D_,process_:Mean]:=Append[First[#][[{1,2}]],process[#[[3]]&/@#]]&/@GatherBy[data3D,#[[{1,2}]]&]

End[ ]


EndPackage[ ]
